
# coding: utf-8

# # Principal Component Analysis
# 
# ### Laurenz Wiskott (2016-09-03)
# 
# In this notebook data is sampled from a Gaussian distribution and then analyzed with principal component analysis.

# ## Imports

# In[ ]:


import numpy as np
import matplotlib.pyplot as plt
import math as m


# ## Generate data

# In[ ]:


numDat = 1000                               # Number of data points. Play with this.
dimDat = 2                                  # Dimensionality of data point. Do not change!


# The data is generated by first sampling from an isotropic Gaussian distribution with zero mean and unit variance in all directions (whitened data) and then transforming the data by a random matrix and random mean. Coefficients of mean vector and transformation matrix are drawn from a normal distribution with variance one.

# In[ ]:


dataInit = np.random.normal(0,1,(dimDat,numDat))        # Initialize data with normal distribution (approximately whitened data).

meanInit = np.random.normal(0,1,dimDat)                 # Initialize mean from normal distribution.
trafo = np.random.normal(0,1,(dimDat,dimDat))           # Create random transformation matrix.
data = np.dot(trafo,dataInit) + meanInit.reshape((2,1)) # Transform initial data.

print "\nInitial data:\n", dataInit
print "\nInitial mean:\n", meanInit
print "\nTransformation:\n", trafo
print "\nData (transformed):\n", data


# ## Calculate mean and principal components of data
# 
# Hints: `np.average`, `np.dot`, `np.transpose`, and `np.linalg.eig` might be useful here.  Use `print np.<name>?` to learn more about `<name>`.

# In[ ]:


# np.average?


# In[ ]:


## EXERCISE 1

# Add missing code here.
# ...
# dataMean = ... # Data mean as 2 dimensional vector.
# pcVec = ... # Principal components as column vectors in a 2x2 dimensional array.
# pcVar = ... # Variances along the principal components as components of a 2 dimensional vector.

# EXERCISE END

print "\nData mean:\n", dataMean
print "\nPrincipal components:\n", pcVec
print "\nVariances:\n", pcVar


# ## Calculate expected mean and principal components from transformation matrix and initial mean
# 
# Since the initial data has zero mean and variance one in all directions, it is possible to predict the data mean and principal components.

# ### Exercise 2
# 
# Figure out analytically how to estimate the data mean and principal components of the data from the transformation matrix and initial mean.

# In[ ]:


## EXERCISE 3

# Add missing code here.
# ...
# meanExp = ... # Expected data mean as 2 dimensional vector.
# pcVecExp = ... # Expected principal components as column vectors in a 2x2 dimensional array.
# pcVarExp = ... # Expected variances along the principal components as components of a 2 dimensional vector.

# EXERCISE END

print "\nExpected data mean:\n", meanExp
print "\nExpected principal components:\n", pcVecExp
print "\nExpected variances:\n", pcVarExp


# # Plot data and eigenvectors scaled by standard deviation and centered at mean
# 
# The following plot shows the data points in black, the true mean offset vector in yellow, the data mean vector in blue on top, the expected eigenvectors in green, and the data eigenvectors in red on top.  Since the polarity of the eigenvectors is arbitrary, they are plotted $\pm$.

# In[ ]:


fig = plt.figure(figsize=(12,12))
ax = fig.add_subplot(111)
pcStd = np.array([m.sqrt(x) for x in pcVar]) # Convert variances to standard deviations, ...
pcStdExp = np.array([m.sqrt(x) for x in pcVarExp]) # ... also for the expected values.
plotRange = 3*np.max(pcStd) # Take the largest standard deviation for the plot range.
ax.set(xlim=[dataMean[0]-plotRange, dataMean[0]+plotRange], ylim=[dataMean[1]-plotRange, dataMean[1]+plotRange], title='Random Vectors', xlabel='First Component', ylabel='Second Component')
ax.scatter(data[0],data[1],s=3) # Plot data points.
ax.plot([0,meanInit[0]],[0,meanInit[1]],
        'yo-',markersize=15,linewidth=5) # Plot ground truth mean offset of the data in yellow.
for component in range(0,2):
    ax.plot([meanExp[0]-pcVecExp[0,component]*pcStdExp[component],meanExp[0]+pcVecExp[0,component]*pcStdExp[component]],[meanExp[1]-pcVecExp[1,component]*pcStdExp[component],meanExp[1]+pcVecExp[1,component]*pcStdExp[component]],
            'go-',markersize=15,linewidth=5) # Plot expected eigenvectors scaled by standard deviation in green, ...
ax.plot([0,dataMean[0]],[0,dataMean[1]],
        'bo-',markersize=15,linewidth=5) # Plot data mean offset of the mixed data in blue.
for component in range(0,2):
    ax.plot([dataMean[0]-pcVec[0,component]*pcStd[component],dataMean[0]+pcVec[0,component]*pcStd[component]],[dataMean[1]-pcVec[1,component]*pcStd[component],dataMean[1]+pcVec[1,component]*pcStd[component]],
            'ro-',markersize=15,linewidth=5) # ... and data eigenvectors in red on top.
plt.show()


# ## Whiten the data
# 
# Knowing the mean, principal components and variances of the data allows us to whiten the data, i.e. normalize it such that it has zero mean and unit variance in all directions.

# In[ ]:


## EXERCISE 4

# Add missing code here.
# ...
# whiteningMatrix = ... # Whitening matrix as a 2 x 2 dimensional array.
# whitenedData = ...    # Whitened data as a 2 x numDat dimensional array.

# EXERCISE END

print "\nWhitening matrix:\n", whiteningMatrix
print "\nWhitened data:\n", whitenedData


# In[ ]:


fig = plt.figure(figsize=(8,8))
ax = fig.add_subplot(111)
plotRange = 4
ax.set(xlim=[-plotRange,+plotRange], ylim=[-plotRange,+plotRange], title='Whitened Data', xlabel='First Component', ylabel='Second Component')
ax.scatter(whitenedData[0],whitenedData[1],s=3) # Plot data points.
plt.show()


# ## Exercise 5
# 
# Vary the number of data points and observe how much the solution depends on it.

# #####  This work is licensed under a <a rel="license" href="http://creativecommons.org/licenses/by/4.0/">Creative Commons Attribution 4.0 International License</a>.
